using ImmichToSlideshow.Models;
using ImmichToSlideshow.Models.Immich;
using Microsoft.AspNetCore.Mvc;
using Npgsql;
using System.Collections.ObjectModel;
using System.Data;
using System.Text;
using System.Text.Json;

namespace ImmichToSlideshow.Services;

public class AssetService(ILogger<Program> logger, AppSettings appSettings)
{

    private readonly ILogger<Program> _Logger = logger;
    private readonly AppSettings _AppSettings = appSettings;

    private static int? ExecuteNonQuery(string connectionString, string commandText, ReadOnlyCollection<NpgsqlParameter> npgsqlParameters)
    {
        int? result;
        if (string.IsNullOrEmpty(connectionString))
            result = null;
        else
        {
            using NpgsqlConnection npgsqlConnection = new(connectionString);
            npgsqlConnection.Open();
            using NpgsqlCommand npgsqlCommand = new(commandText, npgsqlConnection);
            npgsqlCommand.Parameters.AddRange(npgsqlParameters.ToArray());
            result = npgsqlCommand.ExecuteNonQuery();
        }
        return result;
    }

    private static StringBuilder GetForJsonPath(string connectionString, string commandText, ReadOnlyCollection<NpgsqlParameter> npgsqlParameters)
    {
        StringBuilder result = new();
        using NpgsqlConnection npgsqlConnection = new(connectionString);
        npgsqlConnection.Open();
        using NpgsqlCommand npgsqlCommand = new(commandText, npgsqlConnection);
        npgsqlCommand.Parameters.AddRange(npgsqlParameters.ToArray());
        NpgsqlDataReader npgsqlDataReader = npgsqlCommand.ExecuteReader(CommandBehavior.SequentialAccess);
        while (npgsqlDataReader.Read())
            _ = result.Append(npgsqlDataReader.GetString(0));
        return result;
    }

    public string? GetColumns()
    {
        string result;
        NpgsqlParameter[] npgsqlParameters = [];
        string commandText = CommandText.GetColumns();
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/.jsonl", result);
        return result;
    }

    public string? GetOwnerIds()
    {
        string result;
        NpgsqlParameter[] npgsqlParameters = [];
        string commandText = CommandText.GetOwnerIdActiveImage();
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/.jsonl", result);
        return result;
    }

    public string? GetAssets(Guid ownerId)
    {
        string result;
        NpgsqlParameter[] npgsqlParameters = [new NpgsqlParameter(nameof(ownerId), ownerId)];
        string commandText = CommandText.GetAssetActiveImagePreviewNotDuplicate(_AppSettings.FilterTags);
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/assets.jsonl", result);
        return result;
    }

    public ReadOnlyCollection<string>? GetRandomPaths(Guid ownerId)
    {
        string[]? results;
        NpgsqlParameter[] npgsqlParameters = [new NpgsqlParameter(nameof(ownerId), ownerId)];
        string commandText = CommandText.GetAssetActiveImagePreviewNotDuplicate(_AppSettings.FilterTags);
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        string json = stringBuilder.ToString();
        Random random = new();
        string ownerIdValue = ownerId.ToString();
        Asset[]? assets = JsonSerializer.Deserialize(json, AssetCollectionSourceGenerationContext.Default.AssetArray);
        results = assets is null ? null : (from l in assets orderby random.NextSingle() select l.Path.Split(ownerIdValue)[1]).ToArray();
        return results?.AsReadOnly();
    }

    public Tag[]? GetArchivedTag(AppSettings appSettings)
    {
        Tag[] results;
        NpgsqlParameter[] npgsqlParameters = [];
        string commandText = CommandText.GetArchivedTag(appSettings.ArchivedTag);
        StringBuilder stringBuilder = GetForJsonPath(appSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
        string json = stringBuilder.ToString();
        results = JsonSerializer.Deserialize(json, TagCollectionSourceGenerationContext.Default.TagArray);
        return results;
    }

    public string? GetArchivedTag()
    {
        string result;
        Tag[]? tags = GetArchivedTag(_AppSettings);
        result = tags is null || tags.Length != 1 ? null : tags[0].Id;
        return result;
    }

    public string? SaveRandomPaths(Guid ownerId)
    {
        string? results;
        DateTime dateTime = DateTime.Now;
        string tomorrow = dateTime.AddDays(1).ToString("MM-dd");
        if (Directory.Exists(_AppSettings.RandomResultsDirectory))
            _ = Directory.CreateDirectory(_AppSettings.RandomResultsDirectory);
        FileInfo fileInfo = new(Path.Combine(_AppSettings.RandomResultsDirectory, $"{tomorrow}.json"));
        if (fileInfo.Exists && fileInfo.CreationTime > dateTime.AddDays(_AppSettings.AddDays))
            results = null;
        else
        {
            _Logger.LogDebug("Writing <{FullName}>", fileInfo.FullName);
            ReadOnlyCollection<string>? paths = GetRandomPaths(ownerId);
            if (paths is null)
                results = null;
            else
            {
                _Logger.LogInformation("{count} path(s)", paths.Count.ToString());
                results = JsonSerializer.Serialize(paths);
                File.WriteAllText(fileInfo.FullName, results);
            }
        }
        return results;
    }

    private record Record(string Source, string Destination);

    public ReadOnlyCollection<string> SyncImmich(Guid ownerId)
    {
        List<string> results = [];
        Record record;
        List<Record> records = [];
        if (Directory.Exists(_AppSettings.SyncDirectory))
            _ = Directory.CreateDirectory(_AppSettings.SyncDirectory);
        int syncLength = _AppSettings.SyncDirectory.Length;
        string[] syncFiles = Directory.GetFiles(_AppSettings.SyncDirectory, "*", SearchOption.AllDirectories);
        string[] syncCheck = syncFiles.Select(l => l[syncLength..]).ToArray();
        if (Directory.Exists(_AppSettings.ImmichUploadDirectory))
            _ = Directory.CreateDirectory(_AppSettings.ImmichUploadDirectory);
        int immichUploadLength = _AppSettings.ImmichUploadDirectory.Length;
        string[] immichUploadFiles = Directory.GetFiles(_AppSettings.ImmichUploadDirectory, "*", SearchOption.AllDirectories);
        string[] immichUploadCheck = immichUploadFiles.Select(l => l[immichUploadLength..]).ToArray();
        for (int i = 0; i < immichUploadFiles.Length; i++)
        {
            if (syncCheck.Contains(immichUploadCheck[i]))
                continue;
            results.Add(immichUploadCheck[i]);
            record = new(immichUploadFiles[i], string.Concat(_AppSettings.SyncDirectory, immichUploadCheck[i]));
            records.Add(record);
        }
        _Logger.LogInformation("{count} file(s)", results.Count.ToString());
        for (int i = 0; i < records.Count; i++)
        {
            record = records[i];
            _Logger.LogDebug("Copying <{source}>", record.Source);
            File.Copy(record.Source, record.Destination);
        }
        return results.AsReadOnly();
    }

    private static string GetDeviceAssetIds(ReadOnlyCollection<Identifier> identifiers, ReadOnlyCollection<string> assets) =>
        $"'{string.Join($"',{Environment.NewLine}'", (from l in identifiers where !assets.Contains($"{l.PaddedId}{l.Extension}") select $"{l.PaddedId}{l.Extension}").ToArray())}'";

    public ReadOnlyCollection<int> SetArchiveImmich(Guid ownerId)
    {
        List<int> results;
        string checkDirectory = Path.Combine(_AppSettings.CurrentResultsDirectory, "B)Metadata", _AppSettings.CurrentCommit, "[]");
        if (!Directory.Exists(checkDirectory))
            results = null;
        else
        {
            string checkFile = Path.Combine(checkDirectory, "!9.json");
            if (!File.Exists(checkFile))
                results = null;
            else
            {
                string json;
                string commandText;
                json = File.ReadAllText(checkFile);
                Identifier[]? identifiers = JsonSerializer.Deserialize<Identifier[]>(json);
                if (identifiers is null || identifiers.Length == 0)
                    results = null;
                else
                {
                    Tag[]? tags = GetArchivedTag(_AppSettings);
                    if (tags is null || tags.Length != 1)
                        results = null;
                    else
                    {
                        List<NpgsqlParameter> npgsqlParameters = [];
                        npgsqlParameters.Add(new NpgsqlParameter(nameof(ownerId), ownerId));
                        commandText = CommandText.GetAssetActiveImagePreviewNotDuplicate(_AppSettings.FilterTags);
                        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
                        json = stringBuilder.ToString();
                        Asset[]? assets = JsonSerializer.Deserialize(json, AssetCollectionSourceGenerationContext.Default.AssetArray);
                        if (assets is null || assets.Length == 0)
                            results = null;
                        else
                        {
                            string tagsId = tags[0].Id;
                            npgsqlParameters.Add(new NpgsqlParameter(nameof(tagsId), tagsId));
                            string[] assetDeviceAssetIds = (from l in assets select l.DeviceAssetId).ToArray();
                            string deviceAssetIds = GetDeviceAssetIds(identifiers.AsReadOnly(), assetDeviceAssetIds.AsReadOnly());
                            commandText = CommandText.SetAssetArchived(deviceAssetIds);
                            int? result = ExecuteNonQuery(_AppSettings.ConnectionString, commandText, npgsqlParameters.AsReadOnly());
                            if (result is null)
                                results = null;
                            else
                                results = [result.Value];
                        }
                    }
                }
            }
        }
        return results?.AsReadOnly();
    }

}